@@ -25,6 +25,7 @@
 
 #include "gl-enums.h"
 #include "gl-eventtoolbar.h"
+#include "gl-eventviewrow.h"
 #include "gl-journal.h"
 #include "gl-util.h"
 
@@ -49,7 +50,7 @@ typedef struct
 G_DEFINE_TYPE_WITH_PRIVATE (GlEventView, gl_event_view, GTK_TYPE_STACK)
 
 static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };
-static const guint N_RESULTS = 50;
+static const gssize N_RESULTS = -1;
 static const gchar DESKTOP_SCHEMA[] = "org.gnome.desktop.interface";
 static const gchar CLOCK_FORMAT[] = "clock-format";
 
@@ -108,25 +109,14 @@ listbox_search_filter_func (GtkListBoxRow *row,
     }
     else
     {
-        gchar *cursor;
         GlJournalResult *result;
 
-        cursor = g_object_get_data (G_OBJECT (row), "cursor");
-
-        if (cursor == NULL)
-        {
-            g_warning ("Error getting cursor from row");
-            goto out;
-        }
-
-        result = gl_journal_query_cursor (priv->journal, cursor);
+        result = gl_event_view_row_get_result (GL_EVENT_VIEW_ROW (row));
 
         if (gl_event_view_search_is_case_sensitive (view))
         {
             if (search_in_result (result, priv->search_text))
             {
-                gl_journal_result_free (priv->journal, result);
-
                 return TRUE;
             }
         }
@@ -164,11 +154,8 @@ listbox_search_filter_func (GtkListBoxRow *row,
             g_free (casefolded.audit_session);
             g_free (casefolded_text);
         }
-
-        gl_journal_result_free (priv->journal, result);
     }
 
-out:
     return FALSE;
 }
 
@@ -178,7 +165,6 @@ on_listbox_row_activated (GtkListBox *listbox,
                           GlEventView *view)
 {
     GlEventViewPrivate *priv;
-    gchar *cursor;
     GlJournalResult *result;
     gchar *time;
     gboolean rtl;
@@ -188,18 +174,11 @@ on_listbox_row_activated (GtkListBox *listbox,
     GtkStack *stack;
 
     priv = gl_event_view_get_instance_private (view);
-    cursor = g_object_get_data (G_OBJECT (row), "cursor");
 
-    if (cursor == NULL)
-    {
-        g_warning ("Error getting cursor from row");
-        return;
-    }
+    result = gl_event_view_row_get_result (GL_EVENT_VIEW_ROW (row));
 
     rtl = gtk_widget_get_default_direction () == GTK_TEXT_DIR_RTL;
 
-    result = gl_journal_query_cursor (priv->journal, cursor);
-
     grid = gtk_grid_new ();
     label = gtk_label_new (result->comm);
     gtk_widget_set_direction (label, GTK_TEXT_DIR_LTR);
@@ -242,556 +221,9 @@ on_listbox_row_activated (GtkListBox *listbox,
     gtk_stack_add_named (stack, grid, "detail");
     gl_event_view_set_mode (view, GL_EVENT_VIEW_MODE_DETAIL);
 
-    gl_journal_result_free (priv->journal, result);
     return;
 }
 
-static void
-on_notify_filter (GlEventView *view,
-                  G_GNUC_UNUSED GParamSpec *pspec,
-                  G_GNUC_UNUSED gpointer user_data)
-{
-    GlEventViewPrivate *priv;
-    GtkStack *stack;
-    GtkWidget *scrolled;
-    GtkWidget *viewport;
-
-    priv = gl_event_view_get_instance_private (view);
-    stack = GTK_STACK (view);
-
-    switch (priv->filter)
-    {
-        case GL_EVENT_VIEW_FILTER_IMPORTANT:
-            gtk_stack_set_visible_child_name (stack, "listbox-important");
-            break;
-        case GL_EVENT_VIEW_FILTER_ALERTS:
-            gtk_stack_set_visible_child_name (stack, "listbox-alerts");
-            break;
-        case GL_EVENT_VIEW_FILTER_STARRED:
-            gtk_stack_set_visible_child_name (stack, "listbox-starred");
-            break;
-        case GL_EVENT_VIEW_FILTER_ALL:
-            gtk_stack_set_visible_child_name (stack, "listbox-all");
-            break;
-        case GL_EVENT_VIEW_FILTER_APPLICATIONS:
-            gtk_stack_set_visible_child_name (stack, "listbox-applications");
-            break;
-        case GL_EVENT_VIEW_FILTER_SYSTEM:
-            gtk_stack_set_visible_child_name (stack, "listbox-system");
-            break;
-        case GL_EVENT_VIEW_FILTER_HARDWARE:
-            gtk_stack_set_visible_child_name (stack, "listbox-hardware");
-            break;
-        case GL_EVENT_VIEW_FILTER_SECURITY:
-            gtk_stack_set_visible_child_name (stack, "listbox-security");
-            break;
-        case GL_EVENT_VIEW_FILTER_UPDATES:
-            gtk_stack_set_visible_child_name (stack, "listbox-updates");
-            break;
-        case GL_EVENT_VIEW_FILTER_USAGE:
-            gtk_stack_set_visible_child_name (stack, "listbox-usage");
-            break;
-        default:
-            break;
-    }
-
-    scrolled = gtk_stack_get_visible_child (stack);
-    viewport = gtk_bin_get_child (GTK_BIN (scrolled));
-    priv->active_listbox = GTK_LIST_BOX (gtk_bin_get_child (GTK_BIN (viewport)));
-
-    gl_event_view_set_mode (view, GL_EVENT_VIEW_MODE_LIST);
-}
-
-static void
-on_notify_mode (GlEventView *view,
-                GParamSpec *pspec,
-                gpointer user_data)
-{
-    GlEventViewPrivate *priv;
-    GtkStack *stack;
-    GtkWidget *toplevel;
-
-    priv = gl_event_view_get_instance_private (view);
-    stack = GTK_STACK (view);
-
-    switch (priv->mode)
-    {
-        case GL_EVENT_VIEW_MODE_LIST:
-            {
-                GtkContainer *container;
-                GList *children;
-                GList *l;
-                GtkWidget *viewport;
-                GtkWidget *scrolled_window;
-
-                container = GTK_CONTAINER (stack);
-                children = gtk_container_get_children (container);
-
-                for (l = children; l != NULL; l = g_list_next (l))
-                {
-                    GtkWidget *child;
-                    gchar *name;
-
-                    child = (GtkWidget *)l->data;
-                    gtk_container_child_get (container, child, "name", &name,
-                                             NULL);
-
-                    if (g_strcmp0 (name, "detail") == 0)
-                    {
-                        gtk_container_remove (container, child);
-
-                        g_free (name);
-                        break;
-                    }
-
-                    g_free (name);
-                }
-
-                g_list_free (children);
-
-                viewport = gtk_widget_get_parent (GTK_WIDGET (priv->active_listbox));
-                scrolled_window = gtk_widget_get_parent (viewport);
-                gtk_stack_set_visible_child (stack, scrolled_window);
-            }
-            break;
-        case GL_EVENT_VIEW_MODE_DETAIL:
-            gtk_stack_set_visible_child_name (stack, "detail");
-            break;
-        default:
-            g_assert_not_reached ();
-            break;
-    }
-
-    toplevel = gtk_widget_get_toplevel (GTK_WIDGET (view));
-
-    if (gtk_widget_is_toplevel (toplevel))
-    {
-        GAction *mode;
-        GEnumClass *eclass;
-        GEnumValue *evalue;
-
-        mode = g_action_map_lookup_action (G_ACTION_MAP (toplevel), "view-mode");
-        eclass = g_type_class_ref (GL_TYPE_EVENT_VIEW_MODE);
-        evalue = g_enum_get_value (eclass, priv->mode);
-
-        g_action_activate (mode, g_variant_new_string (evalue->value_nick));
-
-        g_type_class_unref (eclass);
-    }
-    else
-    {
-        g_debug ("Widget not in toplevel window, not switching toolbar mode");
-    }
-}
-
-static void
-gl_event_view_finalize (GObject *object)
-{
-    GlEventView *view = GL_EVENT_VIEW (object);
-    GlEventViewPrivate *priv = gl_event_view_get_instance_private (view);
-
-    g_clear_pointer (&priv->search_text, g_free);
-}
-
-static void
-gl_event_view_get_property (GObject *object,
-                            guint prop_id,
-                            GValue *value,
-                            GParamSpec *pspec)
-{
-    GlEventView *view = GL_EVENT_VIEW (object);
-    GlEventViewPrivate *priv = gl_event_view_get_instance_private (view);
-
-    switch (prop_id)
-    {
-        case PROP_FILTER:
-            g_value_set_enum (value, priv->filter);
-            break;
-        case PROP_MODE:
-            g_value_set_enum (value, priv->mode);
-        default:
-            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-            break;
-    }
-}
-
-static void
-gl_event_view_set_property (GObject *object,
-                            guint prop_id,
-                            const GValue *value,
-                            GParamSpec *pspec)
-{
-    GlEventView *view = GL_EVENT_VIEW (object);
-    GlEventViewPrivate *priv = gl_event_view_get_instance_private (view);
-
-    switch (prop_id)
-    {
-        case PROP_FILTER:
-            priv->filter = g_value_get_enum (value);
-            break;
-        case PROP_MODE:
-            priv->mode = g_value_get_enum (value);
-            break;
-        default:
-            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-            break;
-    }
-}
-
-static void
-gl_event_view_class_init (GlEventViewClass *klass)
-{
-    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
-
-    gobject_class->finalize = gl_event_view_finalize;
-    gobject_class->get_property = gl_event_view_get_property;
-    gobject_class->set_property = gl_event_view_set_property;
-
-    obj_properties[PROP_FILTER] = g_param_spec_enum ("filter", "Filter",
-                                                     "Filter events by",
-                                                     GL_TYPE_EVENT_VIEW_FILTER,
-                                                     GL_EVENT_VIEW_FILTER_IMPORTANT,
-                                                     G_PARAM_READWRITE |
-                                                     G_PARAM_STATIC_STRINGS);
-
-    obj_properties[PROP_MODE] = g_param_spec_enum ("mode", "Mode",
-                                                   "Event display mode",
-                                                   GL_TYPE_EVENT_VIEW_MODE,
-                                                   GL_EVENT_VIEW_MODE_LIST,
-                                                   G_PARAM_READWRITE |
-                                                   G_PARAM_STATIC_STRINGS);
-
-    g_object_class_install_properties (gobject_class, N_PROPERTIES,
-                                       obj_properties);
-}
-
-static void
-insert_journal_query_devices (GlEventView *view,
-                              const GlJournalQuery *query,
-                              GtkListBox *listbox)
-{
-    GlEventViewPrivate *priv;
-    GList *results = NULL;
-    GList *l;
-    gsize n_results;
-
-    priv = gl_event_view_get_instance_private (view);
-    results = gl_journal_query (priv->journal, query);
-
-    n_results = g_list_length (results);
-
-    if (n_results != N_RESULTS)
-    {
-        g_debug ("Number of results different than requested");
-    }
-
-    for (l = results; l != NULL; l = g_list_next (l))
-    {
-        GtkWidget *row;
-        GtkStyleContext *context;
-        GtkWidget *grid;
-        GtkWidget *label;
-        gchar *time;
-        gboolean rtl;
-        GtkWidget *image;
-        GlJournalResult result = *(GlJournalResult *)(l->data);
-
-        /* Skip if the log entry does not refer to a hardware device. */
-        if (result.kernel_device == NULL)
-        {
-            continue;
-        }
-
-        rtl = (gtk_widget_get_default_direction () == GTK_TEXT_DIR_RTL);
-
-        row = gtk_list_box_row_new ();
-        context = gtk_widget_get_style_context (GTK_WIDGET (row));
-        gtk_style_context_add_class (context, "event");
-        g_object_set_data_full (G_OBJECT (row), "cursor",
-                                g_strdup (result.cursor), g_free);
-        grid = gtk_grid_new ();
-        gtk_grid_set_column_spacing (GTK_GRID (grid), 6);
-        gtk_container_add (GTK_CONTAINER (row), grid);
-
-        label = gtk_label_new (result.message);
-        gtk_widget_set_direction (label, GTK_TEXT_DIR_LTR);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "event-monospace");
-        gtk_widget_set_hexpand (label, TRUE);
-        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-        gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
-        gtk_grid_attach (GTK_GRID (grid), label, 0, 1, 1, 1);
-
-        time = gl_util_timestamp_to_display (result.timestamp,
-                                             priv->clock_format);
-        label = gtk_label_new (time);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "dim-label");
-        gtk_style_context_add_class (context, "event-time");
-        gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
-        gtk_grid_attach (GTK_GRID (grid), label, rtl ? 1 : 0, 0, 1, 1);
-
-        image = gtk_image_new_from_icon_name (rtl ? "go-next-rtl-symbolic"
-                                                  : "go-next-symbolic",
-                                              GTK_ICON_SIZE_MENU);
-        gtk_grid_attach (GTK_GRID (grid), image, 1, 0, 1, 2);
-
-        gtk_container_add (GTK_CONTAINER (listbox), row);
-
-        g_free (time);
-    }
-
-    gl_journal_results_free (priv->journal, results);
-}
-
-static void
-insert_journal_query_security (GlEventView *view,
-                               const GlJournalQuery *query,
-                               GtkListBox *listbox)
-{
-    GlEventViewPrivate *priv;
-    GList *results = NULL;
-    GList *l;
-    gsize n_results;
-
-    priv = gl_event_view_get_instance_private (view);
-    results = gl_journal_query (priv->journal, query);
-
-    n_results = g_list_length (results);
-
-    if (n_results != N_RESULTS)
-    {
-        g_debug ("Number of results different than requested");
-    }
-
-    for (l = results; l != NULL; l = g_list_next (l))
-    {
-        GtkWidget *row;
-        GtkStyleContext *context;
-        GtkWidget *grid;
-        GtkWidget *label;
-        gchar *markup;
-        gchar *time;
-        gboolean rtl;
-        GtkWidget *image;
-        GlJournalResult result = *(GlJournalResult *)(l->data);
-
-        /* Skip if the journal entry does not have an associated audit
-         * session. */
-        if (result.audit_session == NULL)
-        {
-            continue;
-        }
-
-        rtl = (gtk_widget_get_default_direction () == GTK_TEXT_DIR_RTL);
-
-        row = gtk_list_box_row_new ();
-        gtk_widget_set_direction (row, GTK_TEXT_DIR_LTR);
-        context = gtk_widget_get_style_context (GTK_WIDGET (row));
-        gtk_style_context_add_class (context, "event");
-        g_object_set_data_full (G_OBJECT (row), "cursor",
-                                g_strdup (result.cursor), g_free);
-        grid = gtk_grid_new ();
-        gtk_grid_set_column_spacing (GTK_GRID (grid), 6);
-        gtk_container_add (GTK_CONTAINER (row), grid);
-
-        markup = g_markup_printf_escaped ("<b>%s</b>",
-                                          result.comm ? result.comm : "");
-        label = gtk_label_new (NULL);
-        gtk_widget_set_hexpand (label, TRUE);
-        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-        gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
-        gtk_label_set_markup (GTK_LABEL (label), markup);
-        g_free (markup);
-        gtk_grid_attach (GTK_GRID (grid), label, rtl ? 1 : 0, 0, 1, 1);
-
-        label = gtk_label_new (result.message);
-        gtk_widget_set_direction (label, GTK_TEXT_DIR_LTR);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "event-monospace");
-        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-        gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
-        gtk_grid_attach (GTK_GRID (grid), label, 0, 1, 2, 1);
-
-        time = gl_util_timestamp_to_display (result.timestamp,
-                                             priv->clock_format);
-        label = gtk_label_new (time);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "dim-label");
-        gtk_style_context_add_class (context, "event-time");
-        gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
-        gtk_grid_attach (GTK_GRID (grid), label, rtl ? 0 : 1, 0, 1, 1);
-
-        image = gtk_image_new_from_icon_name (rtl ? "go-next-rtl-symbolic"
-                                                  : "go-next-symbolic",
-                                              GTK_ICON_SIZE_MENU);
-        gtk_grid_attach (GTK_GRID (grid), image, 2, 0, 1, 2);
-
-        gtk_container_add (GTK_CONTAINER (listbox), row);
-
-        g_free (time);
-    }
-
-    gl_journal_results_free (priv->journal, results);
-}
-
-static void
-insert_journal_query_simple (GlEventView *view,
-                             const GlJournalQuery *query,
-                             GtkListBox *listbox)
-{
-    GlEventViewPrivate *priv;
-    GList *results = NULL;
-    GList *l;
-    gsize n_results;
-
-    priv = gl_event_view_get_instance_private (view);
-    results = gl_journal_query (priv->journal, query);
-
-    n_results = g_list_length (results);
-
-    if (n_results != N_RESULTS)
-    {
-        g_debug ("Number of results different than requested");
-    }
-
-    for (l = results; l != NULL; l = g_list_next (l))
-    {
-        GtkWidget *row;
-        GtkStyleContext *context;
-        GtkWidget *grid;
-        GtkWidget *label;
-        gchar *time;
-        gboolean rtl;
-        GtkWidget *image;
-        GlJournalResult result = *(GlJournalResult *)(l->data);
-
-        rtl = (gtk_widget_get_default_direction () == GTK_TEXT_DIR_RTL);
-
-        row = gtk_list_box_row_new ();
-        context = gtk_widget_get_style_context (GTK_WIDGET (row));
-        gtk_style_context_add_class (context, "event");
-        g_object_set_data_full (G_OBJECT (row), "cursor",
-                                g_strdup (result.cursor), g_free);
-        grid = gtk_grid_new ();
-        gtk_grid_set_column_spacing (GTK_GRID (grid), 6);
-        gtk_container_add (GTK_CONTAINER (row), grid);
-
-        label = gtk_label_new (result.message);
-        gtk_widget_set_direction (label, GTK_TEXT_DIR_LTR);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "event-monospace");
-        gtk_widget_set_hexpand (label, TRUE);
-        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-        gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
-        gtk_grid_attach (GTK_GRID (grid), label, 0, 1, 1, 1);
-
-        time = gl_util_timestamp_to_display (result.timestamp,
-                                             priv->clock_format);
-        label = gtk_label_new (time);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "dim-label");
-        gtk_style_context_add_class (context, "event-time");
-        gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
-        gtk_grid_attach (GTK_GRID (grid), label, rtl ? 1 : 0, 0, 1, 1);
-
-        image = gtk_image_new_from_icon_name (rtl ? "go-next-rtl-symbolic"
-                                                  : "go-next-symbolic",
-                                              GTK_ICON_SIZE_MENU);
-        gtk_grid_attach (GTK_GRID (grid), image, 1, 0, 1, 2);
-
-        gtk_container_add (GTK_CONTAINER (listbox), row);
-
-        g_free (time);
-    }
-
-    gl_journal_results_free (priv->journal, results);
-}
-
-static void
-insert_journal_query_cmdline (GlEventView *view,
-                              const GlJournalQuery *query,
-                              GtkListBox *listbox)
-{
-    GlEventViewPrivate *priv;
-    GList *results = NULL;
-    GList *l;
-    gsize n_results;
-
-    priv = gl_event_view_get_instance_private (view);
-    results = gl_journal_query (priv->journal, query);
-
-    n_results = g_list_length (results);
-
-    if (n_results != N_RESULTS)
-    {
-        g_debug ("Number of results different than requested");
-    }
-
-    for (l = results; l != NULL; l = g_list_next (l))
-    {
-        GtkWidget *row;
-        GtkStyleContext *context;
-        GtkWidget *grid;
-        gchar *markup;
-        GtkWidget *label;
-        gchar *time;
-        gboolean rtl;
-        GtkWidget *image;
-        GlJournalResult result = *(GlJournalResult *)(l->data);
-
-        rtl = (gtk_widget_get_default_direction () == GTK_TEXT_DIR_RTL);
-
-        row = gtk_list_box_row_new ();
-        context = gtk_widget_get_style_context (GTK_WIDGET (row));
-        gtk_style_context_add_class (context, "event");
-        g_object_set_data_full (G_OBJECT (row), "cursor",
-                                g_strdup (result.cursor), g_free);
-        grid = gtk_grid_new ();
-        gtk_grid_set_column_spacing (GTK_GRID (grid), 6);
-        gtk_container_add (GTK_CONTAINER (row), grid);
-
-        markup = g_markup_printf_escaped ("<b>%s</b>",
-                                          result.comm ? result.comm : "");
-        label = gtk_label_new (NULL);
-        gtk_widget_set_direction (label, GTK_TEXT_DIR_LTR);
-        gtk_widget_set_hexpand (label, TRUE);
-        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-        gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
-        gtk_label_set_markup (GTK_LABEL (label), markup);
-        g_free (markup);
-        gtk_grid_attach (GTK_GRID (grid), label, rtl ? 1 : 0, 0, 1, 1);
-
-        label = gtk_label_new (result.message);
-        gtk_widget_set_direction (label, GTK_TEXT_DIR_LTR);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "event-monospace");
-        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-        gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
-        gtk_grid_attach (GTK_GRID (grid), label, 0, 1, 2, 1);
-
-        time = gl_util_timestamp_to_display (result.timestamp,
-                                             priv->clock_format);
-        label = gtk_label_new (time);
-        context = gtk_widget_get_style_context (GTK_WIDGET (label));
-        gtk_style_context_add_class (context, "dim-label");
-        gtk_style_context_add_class (context, "event-time");
-        gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
-        gtk_grid_attach (GTK_GRID (grid), label, rtl ? 0 : 1, 0, 1, 1);
-
-        image = gtk_image_new_from_icon_name (rtl ? "go-next-rtl-symbolic"
-                                                  : "go-next-symbolic",
-                                              GTK_ICON_SIZE_MENU);
-        gtk_grid_attach (GTK_GRID (grid), image, 2, 0, 1, 2);
-
-        gtk_container_add (GTK_CONTAINER (listbox), row);
-
-        g_free (time);
-    }
-
-    gl_journal_results_free (priv->journal, results);
-}
-
 static GtkWidget *
 gl_event_view_create_empty (G_GNUC_UNUSED GlEventView *view)
 {
@@ -803,11 +235,15 @@ gl_event_view_create_empty (G_GNUC_UNUSED GlEventView *view)
 
     box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 12);
     gtk_widget_set_halign (box, GTK_ALIGN_CENTER);
+    gtk_widget_set_hexpand (box, TRUE);
     gtk_widget_set_valign (box, GTK_ALIGN_CENTER);
+    gtk_widget_set_vexpand (box, TRUE);
     context = gtk_widget_get_style_context (box);
     gtk_style_context_add_class (context, "dim-label");
 
     image = gtk_image_new_from_icon_name ("action-unavailable-symbolic", 0);
+    context = gtk_widget_get_style_context (image);
+    gtk_style_context_add_class (context, "dim-label");
     gtk_image_set_pixel_size (GTK_IMAGE (image), 128);
     gtk_container_add (GTK_CONTAINER (box), image);
 
@@ -842,6 +278,151 @@ gl_event_view_list_box_new (GlEventView *view)
     return listbox;
 }
 
+static void
+insert_journal_query_devices (GlEventView *view,
+                              const GlJournalQuery *query,
+                              GtkListBox *listbox)
+{
+    GlEventViewPrivate *priv;
+    GList *results = NULL;
+    GList *l;
+    gsize n_results;
+
+    priv = gl_event_view_get_instance_private (view);
+    results = gl_journal_query (priv->journal, query);
+
+    n_results = g_list_length (results);
+
+    if ((n_results != -1) && (n_results != N_RESULTS))
+    {
+        g_debug ("Number of results different than requested");
+    }
+
+    for (l = results; l != NULL; l = g_list_next (l))
+    {
+        GtkWidget *row;
+        GlJournalResult *result = (GlJournalResult *)l->data;
+
+        /* Skip if the log entry does not refer to a hardware device. */
+        if (result->kernel_device == NULL)
+        {
+            continue;
+        }
+
+        row = gl_event_view_row_new (result, GL_EVENT_VIEW_ROW_STYLE_SIMPLE,
+                                     priv->clock_format);
+        gtk_container_add (GTK_CONTAINER (listbox), row);
+    }
+
+    gl_journal_results_free (results);
+}
+
+static void
+insert_journal_query_security (GlEventView *view,
+                               const GlJournalQuery *query,
+                               GtkListBox *listbox)
+{
+    GlEventViewPrivate *priv;
+    GList *results = NULL;
+    GList *l;
+    gsize n_results;
+
+    priv = gl_event_view_get_instance_private (view);
+    results = gl_journal_query (priv->journal, query);
+
+    n_results = g_list_length (results);
+
+    if ((n_results != -1) && (n_results != N_RESULTS))
+    {
+        g_debug ("Number of results different than requested");
+    }
+
+    for (l = results; l != NULL; l = g_list_next (l))
+    {
+        GtkWidget *row;
+        GlJournalResult *result = (GlJournalResult *)l->data;
+
+        /* Skip if the journal entry does not have an associated audit
+         * session. */
+        if (result->audit_session == NULL)
+        {
+            continue;
+        }
+
+        row = gl_event_view_row_new (result, GL_EVENT_VIEW_ROW_STYLE_CMDLINE,
+                                     priv->clock_format);
+        gtk_container_add (GTK_CONTAINER (listbox), row);
+    }
+
+    gl_journal_results_free (results);
+}
+
+static void
+insert_journal_query_simple (GlEventView *view,
+                             const GlJournalQuery *query,
+                             GtkListBox *listbox)
+{
+    GlEventViewPrivate *priv;
+    GList *results = NULL;
+    GList *l;
+    gsize n_results;
+
+    priv = gl_event_view_get_instance_private (view);
+    results = gl_journal_query (priv->journal, query);
+
+    n_results = g_list_length (results);
+
+    if ((n_results != -1) && (n_results != N_RESULTS))
+    {
+        g_debug ("Number of results different than requested");
+    }
+
+    for (l = results; l != NULL; l = g_list_next (l))
+    {
+        GtkWidget *row;
+
+        row = gl_event_view_row_new ((GlJournalResult *)l->data,
+                                     GL_EVENT_VIEW_ROW_STYLE_SIMPLE,
+                                     priv->clock_format);
+        gtk_container_add (GTK_CONTAINER (listbox), row);
+    }
+
+    gl_journal_results_free (results);
+}
+
+static void
+insert_journal_query_cmdline (GlEventView *view,
+                              const GlJournalQuery *query,
+                              GtkListBox *listbox)
+{
+    GlEventViewPrivate *priv;
+    GList *results = NULL;
+    GList *l;
+    gsize n_results;
+
+    priv = gl_event_view_get_instance_private (view);
+    results = gl_journal_query (priv->journal, query);
+
+    n_results = g_list_length (results);
+
+    if ((n_results != -1) && (n_results != N_RESULTS))
+    {
+        g_debug ("Number of results different than requested");
+    }
+
+    for (l = results; l != NULL; l = g_list_next (l))
+    {
+        GtkWidget *row;
+
+        row = gl_event_view_row_new ((GlJournalResult *)l->data,
+                                     GL_EVENT_VIEW_ROW_STYLE_CMDLINE,
+                                     priv->clock_format);
+        gtk_container_add (GTK_CONTAINER (listbox), row);
+    }
+
+    gl_journal_results_free (results);
+}
+
 static void
 gl_event_view_add_listbox_important (GlEventView *view)
 {
@@ -1027,6 +608,243 @@ gl_event_view_add_listbox_usage (GlEventView *view)
     gtk_stack_add_named (GTK_STACK (view), label, "listbox-usage");
 }
 
+static void
+on_notify_filter (GlEventView *view,
+                  G_GNUC_UNUSED GParamSpec *pspec,
+                  G_GNUC_UNUSED gpointer user_data)
+{
+    GlEventViewPrivate *priv;
+    GtkStack *stack;
+    GtkWidget *scrolled;
+    GtkWidget *viewport;
+
+    priv = gl_event_view_get_instance_private (view);
+    stack = GTK_STACK (view);
+
+    if (priv->active_listbox)
+    {
+        GtkWidget *child;
+
+        child = gtk_stack_get_visible_child (stack);
+        gtk_widget_destroy (child);
+    }
+
+    switch (priv->filter)
+    {
+        case GL_EVENT_VIEW_FILTER_IMPORTANT:
+            gl_event_view_add_listbox_important (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-important");
+            break;
+        case GL_EVENT_VIEW_FILTER_ALERTS:
+            gl_event_view_add_listbox_alerts (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-alerts");
+            break;
+        case GL_EVENT_VIEW_FILTER_STARRED:
+            gl_event_view_add_listbox_starred (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-starred");
+            break;
+        case GL_EVENT_VIEW_FILTER_ALL:
+            gl_event_view_add_listbox_all (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-all");
+            break;
+        case GL_EVENT_VIEW_FILTER_APPLICATIONS:
+            gl_event_view_add_listbox_applications (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-applications");
+            break;
+        case GL_EVENT_VIEW_FILTER_SYSTEM:
+            gl_event_view_add_listbox_system (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-system");
+            break;
+        case GL_EVENT_VIEW_FILTER_HARDWARE:
+            gl_event_view_add_listbox_hardware (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-hardware");
+            break;
+        case GL_EVENT_VIEW_FILTER_SECURITY:
+            gl_event_view_add_listbox_security (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-security");
+            break;
+        case GL_EVENT_VIEW_FILTER_UPDATES:
+            gl_event_view_add_listbox_updates (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-updates");
+            break;
+        case GL_EVENT_VIEW_FILTER_USAGE:
+            gl_event_view_add_listbox_usage (view);
+            gtk_stack_set_visible_child_name (stack, "listbox-usage");
+            break;
+        default:
+            break;
+    }
+
+    scrolled = gtk_stack_get_visible_child (stack);
+    viewport = gtk_bin_get_child (GTK_BIN (scrolled));
+    priv->active_listbox = GTK_LIST_BOX (gtk_bin_get_child (GTK_BIN (viewport)));
+
+    gl_event_view_set_mode (view, GL_EVENT_VIEW_MODE_LIST);
+}
+
+static void
+on_notify_mode (GlEventView *view,
+                GParamSpec *pspec,
+                gpointer user_data)
+{
+    GlEventViewPrivate *priv;
+    GtkStack *stack;
+    GtkWidget *toplevel;
+
+    priv = gl_event_view_get_instance_private (view);
+    stack = GTK_STACK (view);
+
+    switch (priv->mode)
+    {
+        case GL_EVENT_VIEW_MODE_LIST:
+            {
+                GtkContainer *container;
+                GList *children;
+                GList *l;
+                GtkWidget *viewport;
+                GtkWidget *scrolled_window;
+
+                container = GTK_CONTAINER (stack);
+                children = gtk_container_get_children (container);
+
+                for (l = children; l != NULL; l = g_list_next (l))
+                {
+                    GtkWidget *child;
+                    gchar *name;
+
+                    child = (GtkWidget *)l->data;
+                    gtk_container_child_get (container, child, "name", &name,
+                                             NULL);
+
+                    if (g_strcmp0 (name, "detail") == 0)
+                    {
+                        gtk_container_remove (container, child);
+
+                        g_free (name);
+                        break;
+                    }
+
+                    g_free (name);
+                }
+
+                g_list_free (children);
+
+                viewport = gtk_widget_get_parent (GTK_WIDGET (priv->active_listbox));
+                scrolled_window = gtk_widget_get_parent (viewport);
+                gtk_stack_set_visible_child (stack, scrolled_window);
+            }
+            break;
+        case GL_EVENT_VIEW_MODE_DETAIL:
+            gtk_stack_set_visible_child_name (stack, "detail");
+            break;
+        default:
+            g_assert_not_reached ();
+            break;
+    }
+
+    toplevel = gtk_widget_get_toplevel (GTK_WIDGET (view));
+
+    if (gtk_widget_is_toplevel (toplevel))
+    {
+        GAction *mode;
+        GEnumClass *eclass;
+        GEnumValue *evalue;
+
+        mode = g_action_map_lookup_action (G_ACTION_MAP (toplevel), "view-mode");
+        eclass = g_type_class_ref (GL_TYPE_EVENT_VIEW_MODE);
+        evalue = g_enum_get_value (eclass, priv->mode);
+
+        g_action_activate (mode, g_variant_new_string (evalue->value_nick));
+
+        g_type_class_unref (eclass);
+    }
+    else
+    {
+        g_debug ("Widget not in toplevel window, not switching toolbar mode");
+    }
+}
+
+static void
+gl_event_view_finalize (GObject *object)
+{
+    GlEventView *view = GL_EVENT_VIEW (object);
+    GlEventViewPrivate *priv = gl_event_view_get_instance_private (view);
+
+    g_clear_pointer (&priv->search_text, g_free);
+}
+
+static void
+gl_event_view_get_property (GObject *object,
+                            guint prop_id,
+                            GValue *value,
+                            GParamSpec *pspec)
+{
+    GlEventView *view = GL_EVENT_VIEW (object);
+    GlEventViewPrivate *priv = gl_event_view_get_instance_private (view);
+
+    switch (prop_id)
+    {
+        case PROP_FILTER:
+            g_value_set_enum (value, priv->filter);
+            break;
+        case PROP_MODE:
+            g_value_set_enum (value, priv->mode);
+        default:
+            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+            break;
+    }
+}
+
+static void
+gl_event_view_set_property (GObject *object,
+                            guint prop_id,
+                            const GValue *value,
+                            GParamSpec *pspec)
+{
+    GlEventView *view = GL_EVENT_VIEW (object);
+    GlEventViewPrivate *priv = gl_event_view_get_instance_private (view);
+
+    switch (prop_id)
+    {
+        case PROP_FILTER:
+            priv->filter = g_value_get_enum (value);
+            break;
+        case PROP_MODE:
+            priv->mode = g_value_get_enum (value);
+            break;
+        default:
+            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+            break;
+    }
+}
+
+static void
+gl_event_view_class_init (GlEventViewClass *klass)
+{
+    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
+
+    gobject_class->finalize = gl_event_view_finalize;
+    gobject_class->get_property = gl_event_view_get_property;
+    gobject_class->set_property = gl_event_view_set_property;
+
+    obj_properties[PROP_FILTER] = g_param_spec_enum ("filter", "Filter",
+                                                     "Filter events by",
+                                                     GL_TYPE_EVENT_VIEW_FILTER,
+                                                     GL_EVENT_VIEW_FILTER_IMPORTANT,
+                                                     G_PARAM_READWRITE |
+                                                     G_PARAM_STATIC_STRINGS);
+
+    obj_properties[PROP_MODE] = g_param_spec_enum ("mode", "Mode",
+                                                   "Event display mode",
+                                                   GL_TYPE_EVENT_VIEW_MODE,
+                                                   GL_EVENT_VIEW_MODE_LIST,
+                                                   G_PARAM_READWRITE |
+                                                   G_PARAM_STATIC_STRINGS);
+
+    g_object_class_install_properties (gobject_class, N_PROPERTIES,
+                                       obj_properties);
+}
+
 static void
 gl_event_view_init (GlEventView *view)
 {
@@ -1043,17 +861,6 @@ gl_event_view_init (GlEventView *view)
     priv->clock_format = g_settings_get_enum (settings, CLOCK_FORMAT);
     g_object_unref (settings);
 
-    gl_event_view_add_listbox_important (view);
-    gl_event_view_add_listbox_alerts (view);
-    gl_event_view_add_listbox_starred (view);
-    gl_event_view_add_listbox_all (view);
-    gl_event_view_add_listbox_applications (view);
-    gl_event_view_add_listbox_system (view);
-    gl_event_view_add_listbox_hardware (view);
-    gl_event_view_add_listbox_security (view);
-    gl_event_view_add_listbox_updates (view);
-    gl_event_view_add_listbox_usage (view);
-
     g_signal_connect (view, "notify::filter", G_CALLBACK (on_notify_filter),
                       NULL);
     g_signal_connect (view, "notify::mode", G_CALLBACK (on_notify_mode),
